#!/usr/bin/env python3

import gi
import re

from gi.repository import GLib, Gio, Ide

THRESHOLD_CHOICES = {
    'INFO': Ide.DiagnosticSeverity.NOTE,
    'WARNING': Ide.DiagnosticSeverity.WARNING,
    'ERROR': Ide.DiagnosticSeverity.ERROR,
    'SEVERE': Ide.DiagnosticSeverity.FATAL,
    'NONE': Ide.DiagnosticSeverity.NOTE,
}

class RstcheckDiagnosticProvider(Ide.DiagnosticTool):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_program_name('rstcheck')

    def do_configure_launcher(self, launcher):
        # rstcheck - signifies that stdin will be used
        launcher.push_argv('-')

    def do_populate_diagnostics(self, diagnostics, file, stdout, stderr):
        try:
            if stderr is None or len(stderr) < 1:
                return

            for line in stderr.strip().split('\n'):
                # Example diagnostic text is:
                # '-:4: (WARNING/2) Inline strong start-string without end-string.'
                #
                # And this regex operation turns it into:
                # ['-', '4', 'WARNING', '2', 'Inline strong start-string without end-string.']
                text = re.split('\:([0-9]+)\:\s\(([A-Z]+)\/([0-9]{1})\)\s', line)

                file_name = text[0]
                on_line = int(text[1]) - 1
                warning_level = text[2]
                message = text[4]

                start = Ide.Location.new(file, on_line, 0)
                severity = THRESHOLD_CHOICES[warning_level]
                diagnostic = Ide.Diagnostic.new(severity, message, start)

                diagnostics.add(diagnostic)
        except Exception as e:
            Ide.warning('Failed to analyze reStructuredText content: {}'.format(e))



# __init__.py
#
# Copyright 2017 Matthew Leeds <mleeds@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from os import path

from gi.repository import GObject
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GtkSource
from gi.repository import Ide
from gi.repository import Template

_ = Ide.gettext

class MakeBuildSystemDiscovery(Ide.SimpleBuildSystemDiscovery):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.props.glob = '+(GNUmakefile|makefile|Makefile)'
        self.props.hint = 'make_plugin'
        self.props.priority = 1000

class MakeBuildSystem(Ide.Object, Ide.BuildSystem):
    project_file = GObject.Property(type=Gio.File)
    make_dir = GObject.Property(type=Gio.File)
    run_args = None

    def do_parent_set(self, parent):
        if self.project_file.query_file_type(0, None) == Gio.FileType.DIRECTORY:
            self.make_dir = self.project_file
        else:
            self.make_dir = self.project_file.get_parent()

    def do_get_id(self):
        return 'make'

    def do_get_display_name(self):
        return 'Make'

    def do_get_priority(self):
        return 0

    def do_get_builddir(self, pipeline):
        return self.get_context().ref_workdir().get_path()

    def get_make_dir(self):
        return self.make_dir

class MakePipelineAddin(Ide.Object, Ide.PipelineAddin):
    """
    The MakePipelineAddin registers stages to be executed when various
    phases of the build pipeline are requested.
    """

    def do_load(self, pipeline):
        context = pipeline.get_context()
        build_system = Ide.BuildSystem.from_context(context)

        # Only register stages if we are a makefile project
        if type(build_system) != MakeBuildSystem:
            return

        config = pipeline.get_config()
        runtime = config.get_runtime()

        # If the configuration has set $MAKE, then use it.
        make = config.getenv('MAKE') or "make"

        srcdir = context.ref_workdir().get_path()
        builddir = pipeline.get_builddir()

        # Register the build launcher which will perform the incremental
        # build of the project when the Ide.PipelinePhase.BUILD phase is
        # requested of the pipeline.
        build_launcher = pipeline.create_launcher()
        build_launcher.set_cwd(build_system.get_make_dir().get_path())
        build_launcher.push_argv(make)
        build_args = config.get_args_for_phase(Ide.PipelinePhase.BUILD)
        if build_args:
            build_launcher.push_args(build_args)
        if config.props.parallelism > 0:
            build_launcher.push_argv('-j{}'.format(config.props.parallelism))

        clean_launcher = pipeline.create_launcher()
        clean_launcher.set_cwd(build_system.get_make_dir().get_path())
        clean_launcher.push_argv(make)
        clean_launcher.push_argv('clean')

        build_stage = Ide.PipelineStageLauncher.new(context, build_launcher)
        build_stage.set_name(_("Build project"))
        build_stage.set_clean_launcher(clean_launcher)
        build_stage.connect('query', self._query)
        self.track(pipeline.attach(Ide.PipelinePhase.BUILD, 0, build_stage))

        # Register the install launcher which will perform our
        # "make install" when the Ide.PipelinePhase.INSTALL phase
        # is requested of the pipeline.
        install_launcher = pipeline.create_launcher()
        install_launcher.set_cwd(build_system.get_make_dir().get_path())
        install_launcher.push_argv(make)
        install_launcher.push_argv('install')
        install_args = config.get_args_for_phase(Ide.PipelinePhase.INSTALL)
        if install_args:
            install_launcher.push_args(install_args)

        install_stage = Ide.PipelineStageLauncher.new(context, install_launcher)
        install_stage.set_name(_("Install project"))
        self.track(pipeline.attach(Ide.PipelinePhase.INSTALL, 0, install_stage))

        # Determine what it will take to "make run" for this pipeline
        # and stash it on the build_system for use by the build target.
        # This allows us to run Make projects as long as the makefile
        # has a "run" target.
        build_system.run_args = [make, '-C', build_system.get_make_dir().get_path(), 'run']

    def _query(self, stage, pipeline, targets, cancellable):
        stage.set_completed(False)

class MakeBuildTarget(Ide.Object, Ide.BuildTarget):

    def do_get_install_directory(self):
        return None

    def do_get_name(self):
        return 'make-run'

    def do_get_language(self):
        # Not meaningful, since we have an indirect process.
        return 'make'

    def do_get_argv(self):
        context = self.get_context()
        build_system = Ide.BuildSystem.from_context(context)
        assert type(build_system) == MakeBuildSystem
        return build_system.run_args

    def do_get_priority(self):
        return 0

class MakeBuildTargetProvider(Ide.Object, Ide.BuildTargetProvider):
    """
    The MakeBuildTargetProvider just wraps a "make run" command. If the
    Makefile doesn't have a run target, we'll just fail to execute and the user
    should get the warning in their application output (and can update their
    Makefile appropriately).
    """

    def do_get_targets_async(self, cancellable, callback, data):
        task = Gio.Task.new(self, cancellable, callback)
        task.set_priority(GLib.PRIORITY_LOW)

        context = self.get_context()
        build_system = Ide.BuildSystem.from_context(context)

        if type(build_system) != MakeBuildSystem:
            task.return_error(GLib.Error('Not a make build system',
                                         domain=GLib.quark_to_string(Gio.io_error_quark()),
                                         code=Gio.IOErrorEnum.NOT_SUPPORTED))
            return

        task.targets = [MakeBuildTarget()]
        task.return_boolean(True)

    def do_get_targets_finish(self, result):
        if result.propagate_boolean():
            return result.targets

class MakeProjectTemplateProvider(GObject.Object, Ide.TemplateProvider):
    def do_get_project_templates(self):
        return [SimpleMakefileTemplate()]

class MakeTemplateLocator(Template.TemplateLocator):
    license = None

    def empty(self):
        return Gio.MemoryInputStream()

    def do_locate(self, path):
        if path.startswith('license.'):
            filename = GLib.basename(path)
            manager = GtkSource.LanguageManager.get_default()
            language = manager.guess_language(filename, None)

            if self.license is None or language is None:
                return self.empty()

            header = Ide.language_format_header(language, self.license)
            gbytes = GLib.Bytes(header.encode())

            return Gio.MemoryInputStream.new_from_bytes(gbytes)

        return super().do_locate(self, path)

class MakeTemplateBase(Ide.TemplateBase, Ide.ProjectTemplate):
    def __init__(self, id, name, icon_name, description, languages):
        super().__init__()
        self.id = id
        self.name = name
        self.icon_name = icon_name
        self.description = description
        self.languages = languages

        # Work around https://bugzilla.gnome.org/show_bug.cgi?id=687522
        self.locator = MakeTemplateLocator()
        self.props.locator = self.locator

    def do_get_id(self):
        return self.id

    def do_get_name(self):
        return self.name

    def do_get_icon_name(self):
        return self.icon_name

    def do_get_description(self):
        return self.description

    def do_get_languages(self):
        return self.languages

    def do_expand_async(self, params, cancellable, callback, data):
        self.reset()

        task = Gio.Task.new(self, cancellable, callback)

        if 'language' in params:
            self.language = params['language'].get_string().lower()
        else:
            self.language = 'c'

        if self.language not in ('c', 'c++'):
            task.return_error(GLib.Error('Language %s not supported' % self.language))
            return

        if 'versioning' in params:
            self.versioning = params['versioning'].get_string()
        else:
            self.versioning = ''

        if 'author' in params:
            author_name = params['author'].get_string()
        else:
            author_name = GLib.get_real_name()

        scope = Template.Scope.new()
        scope.get('template').assign_string(self.id)

        name = params['name'].get_string().lower()
        name_ = name.lower().replace('-', '_')
        scope.get('name').assign_string(name)
        scope.get('name_').assign_string(name_)
        scope.get('NAME').assign_string(name.upper().replace('-','_'))

        prefix = name if not name.endswith('-glib') else name[:-5]
        PREFIX = prefix.upper().replace('-','_')
        prefix_ = prefix.lower().replace('-','_')
        PreFix = ''.join([word.capitalize() for word in prefix.lower().split('-')])

        scope.get('prefix').assign_string(prefix)
        scope.get('Prefix').assign_string(prefix.capitalize())
        scope.get('PreFix').assign_string(PreFix)
        scope.get('prefix_').assign_string(prefix_)
        scope.get('PREFIX').assign_string(PREFIX)

        scope.get('language').assign_string(self.language)
        scope.get('author').assign_string(author_name)
        scope.get('exec_name').assign_string(name)

        expands = {
            'prefix': prefix,
            'name_': name_,
            'name': name,
            'exec_name': name,
        }

        if self.language == 'c':
            files = {
                'resources/.gitignore': '.gitignore',
                'resources/Makefile': 'Makefile',
                'resources/main.c': '%(exec_name)s.c',
            }
            self.prepare_files(files)
        elif self.language == 'c++':
            files = {
                'resources/.gitignore': '.gitignore',
                'resources/Makefile': 'Makefile',
                'resources/main.cpp': '%(exec_name)s.cpp',
            }
            self.prepare_files(files)

        modes = {}

        if 'license_full' in params:
            license_full_path = params['license_full'].get_string()
            files[license_full_path] = 'COPYING'

        if 'license_short' in params:
            license_short_path = params['license_short'].get_string()
            license_base = Gio.resources_lookup_data(license_short_path[11:], 0).get_data().decode()
            self.locator.license = license_base

        if 'path' in params:
            dir_path = params['path'].get_string()
        else:
            dir_path = name
        directory = Gio.File.new_for_path(dir_path)
        scope.get('project_path').assign_string(directory.get_path())

        for src, dst in files.items():
            destination = directory.get_child(dst % expands)
            if src.startswith('resource://'):
                self.add_resource(src[11:], destination, scope, modes.get(src, 0))
            else:
                path = os.path.join('/plugins/make_plugin', src)
                self.add_resource(path, destination, scope, modes.get(src, 0))

        self.expand_all_async(cancellable, self.expand_all_cb, task)

    def do_expand_finish(self, result):
        return result.propagate_boolean()

    def expand_all_cb(self, obj, result, task):
        try:
            self.expand_all_finish(result)
            task.return_boolean(True)
        except Exception as exc:
            if isinstance(exc, GLib.Error):
                task.return_error(exc)
            else:
                task.return_error(GLib.Error(repr(exc)))

class SimpleMakefileTemplate(MakeTemplateBase):
    def __init__(self):
        super().__init__(
            'empty-makefile',
            _('Empty Makefile Project'),
            'pattern-cli',
            _('Create a new empty project using a simple Makefile'),
            ['C', 'C++'],
         )

    def do_get_priority(self):
        return 10000

    def prepare_files(self, files):
        pass



#!/usr/bin/env python3

import os
import json
import gi

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Ide

class BlueprintService(Ide.LspService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_program('blueprint-compiler')

    def do_configure_client(self, client):
        client.add_language("blueprint")

    def do_configure_launcher(self, pipeline, launcher):
        launcher.push_argv('lsp')

class BlueprintDiagnosticProvider(Ide.LspDiagnosticProvider, Ide.DiagnosticProvider):
    def do_load(self):
        BlueprintService.bind_client(self)

class BlueprintCompletionProvider(Ide.LspCompletionProvider, Ide.CompletionProvider):
    def do_load(self, context):
        BlueprintService.bind_client(self)

class BlueprintHoverProvider(Ide.LspHoverProvider):
    def do_prepare(self):
        self.props.priority = 100
        BlueprintService.bind_client(self)

class BlueprintCodeActionProvider(Ide.LspCodeActionProvider, Ide.CodeActionProvider):
    def do_load(self):
        BlueprintService.bind_client(self)



#!/usr/bin/env python3

import os
import gi

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Ide

DEV_MODE = os.getenv('DEV_MODE') and True or False

class PhpService(Ide.Object):
    _client = None
    _has_started = False
    _supervisor = None
    _context = None
    notif = None

    @classmethod
    def from_context(klass, context):
        return context.ensure_child_typed(PhpService)

    @GObject.Property(type=Ide.LspClient)
    def client(self):
        return self._client

    @client.setter
    def client(self, value):
        self._client = value
        self.notify('client')

    @staticmethod
    def on_destroy(self):
        if self._supervisor:
            supervisor, self._supervisor = self._supervisor, None
            supervisor.stop()

    def _on_load_configuration(self, data):
        try:
            d = GLib.Variant('a{sv}', {
                "intelephense": GLib.Variant('a{sv}', {
                    "files": GLib.Variant('a{sv}', {
                        "associations": GLib.Variant('as', ["*.php", "*.phtml"]),
                        "exclude": GLib.Variant('as', [])
                    }),
                    "completion": GLib.Variant('a{sv}', {
                        "insertUseDeclaration": GLib.Variant('b', True),
                        "fullyQualifyGlobalConstantsAndFunctions": GLib.Variant('b', False),
                        "triggerParameterHints": GLib.Variant('b', True),
                        "maxItems": GLib.Variant('i', 100)
                    }),
                    "format": GLib.Variant('a{sv}', {
                        "enable": GLib.Variant('b', True)
                    })
                })
            })

            return d
        except Error as e:
            Ide.critical ('On Load Configuration Error: {}'.format(e.message))
            return GLib.Variant ('a{sv}', {})

    def _on_notification(self, client, name, data):
        if name == "indexingStarted":
            if self.notif is not None:
                self.notif.withdraw()

            self.notif = Ide.Notification(
                id='org.gnome.builder.intelephense.indexing',
                title="Intelephense",
                body=_('Indexing php code…'),
                has_progress=True,
                progress_is_imprecise=True,
                progress=0.0)
            self.notif.attach(self._context)
        elif name == "indexingEnded":
            if self.notif is not None:
                self.notif.withdraw()

    def _get_runtime(self):
        config_manager = Ide.ConfigManager.from_context(self._context)
        config = config_manager.get_current()
        return config.get_runtime()

    def _ensure_started(self):
        # To avoid starting the process unconditionally at startup, lazily
        # start it when the first provider tries to bind a client to its
        # :client property.
        if not self._has_started:
            self._has_started = True

            launcher = self._create_launcher()
            launcher.set_clear_env(False)

            # Locate the directory of the project and run intelephense from there
            workdir = self.get_context().ref_workdir()
            launcher.set_cwd(workdir.get_path())

            launcher.push_argv("intelephense")
            launcher.push_argv("--stdio")

            # Spawn our peer process and monitor it for
            # crashes. We may need to restart it occasionally.
            self._supervisor = Ide.SubprocessSupervisor()
            self._supervisor.connect('spawned', self._ls_spawned)
            self._supervisor.set_launcher(launcher)
            self._supervisor.start()

    def _ls_spawned(self, supervisor, subprocess):
        stdin = subprocess.get_stdin_pipe()
        stdout = subprocess.get_stdout_pipe()
        io_stream = Gio.SimpleIOStream.new(stdout, stdin)

        if self._client:
            self._client.stop()
            self._client.destroy()

        self._client = Ide.LspClient.new(io_stream)
        self._client.connect('load-configuration', self._on_load_configuration)
        self._client.connect('notification', self._on_notification)
        self.append(self._client)
        self._client.add_language('php')
        self._client.start()
        self.notify('client')

    def _create_launcher(self):
        flags = Gio.SubprocessFlags.STDIN_PIPE | Gio.SubprocessFlags.STDOUT_PIPE
        if not DEV_MODE:
            flags |= Gio.SubprocessFlags.STDERR_SILENCE
        launcher = Ide.SubprocessLauncher()
        launcher.set_flags(flags)
        return launcher

    @classmethod
    def bind_client(klass, provider):
        context = provider.get_context()
        self = PhpService.from_context(context)
        self._context = context
        self._ensure_started()
        self.connect('destroy', PhpService.on_destroy)
        self.bind_property('client', provider, 'client', GObject.BindingFlags.SYNC_CREATE)

class PhpLspSymbolResolver(Ide.LspSymbolResolver, Ide.SymbolResolver):
    def do_load(self):
        PhpService.bind_client(self)

class PhpLspHoverProvider(Ide.LspHoverProvider):
    def do_prepare(self):
        self.props.category = 'PHP'
        self.props.priority = 300
        PhpService.bind_client(self)

class PhpLspFormatter(Ide.LspFormatter, Ide.Formatter):
    def do_load(self):
        PhpService.bind_client(self)

class PhpLspDiagnosticProvider(Ide.LspDiagnosticProvider, Ide.DiagnosticProvider):
   def do_load(self):
       PhpService.bind_client(self)

class PhpLspCompletionProvider(Ide.LspCompletionProvider, Ide.CompletionProvider):
    def do_load(self, context):
        PhpService.bind_client(self)

class PhpLspHighlighter(Ide.LspHighlighter, Ide.Highlighter):
    def do_load(self):
        PhpService.bind_client(self)

class PhpLspRenameProvider(Ide.LspRenameProvider, Ide.RenameProvider):
    def do_load(self):
        PhpService.bind_client(self)

class PhpLspCodeActionProvider(Ide.LspCodeActionProvider, Ide.CodeActionProvider):
    def do_load(self):
        PhpService.bind_client(self)
